@@ -434,7 +434,8 @@ validate_tls (GtkBuilder *builder, const char *prefix, GError **error)
 	char *tmp;
 	gboolean valid, encrypted = FALSE;
 	GtkWidget *widget;
-	char *str;
+	NMSettingSecretFlags pw_flags;
+	gboolean secrets_required = TRUE;
 
 	tmp = g_strdup_printf ("%s_ca_cert_chooser", prefix);
 	valid = validate_file_chooser (builder, tmp);
@@ -471,14 +472,19 @@ validate_tls (GtkBuilder *builder, const char *prefix, GError **error)
 	}
 
 	/* Encrypted certificates require a password */
-	str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));
-	encrypted = is_encrypted (str);
-	g_free (str);
-	if (encrypted) {
-		tmp = g_strdup_printf ("%s_private_key_password_entry", prefix);
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, tmp));
-		g_free (tmp);
+	tmp = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));
+	encrypted = is_encrypted (tmp);
+	g_free (tmp);
 
+	tmp = g_strdup_printf ("%s_private_key_password_entry", prefix);
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, tmp));
+	g_free (tmp);
+	pw_flags = nma_utils_menu_to_secret_flags (widget);
+	if (   pw_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED
+	    || pw_flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)
+		secrets_required = FALSE;
+
+	if (encrypted && secrets_required) {
 		if (!gtk_entry_get_text_length (GTK_ENTRY (widget))) {
 			g_set_error (error,
 			             OPENVPN_PLUGIN_UI_ERROR,
@@ -1011,6 +1017,15 @@ populate_cipher_combo (GtkComboBox *box, const char *user_cipher)
 	                    TLS_CIPHER_COL_NAME, _("Default"),
 	                    TLS_CIPHER_COL_DEFAULT, TRUE, -1);
 
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    TLS_CIPHER_COL_NAME, "none",
+	                    TLS_CIPHER_COL_DEFAULT, FALSE, -1);
+	if (g_strcmp0 (user_cipher, "none") == 0) {
+		gtk_combo_box_set_active_iter (box, &iter);
+		user_added = TRUE;
+	}
+
 	items = g_strsplit (tmp, "\n", 0);
 	g_free (tmp);
 
@@ -1027,12 +1042,15 @@ populate_cipher_combo (GtkComboBox *box, const char *user_cipher)
 		if (space)
 			*space = '\0';
 
+		if (strcmp (*item, "none") == 0)
+			continue;
+
 		if (strlen (*item)) {
 			gtk_list_store_append (store, &iter);
 			gtk_list_store_set (store, &iter,
 			                    TLS_CIPHER_COL_NAME, *item,
 			                    TLS_CIPHER_COL_DEFAULT, FALSE, -1);
-			if (user_cipher && !strcmp (*item, user_cipher)) {
+			if (!user_added && user_cipher && !strcmp (*item, user_cipher)) {
 				gtk_combo_box_set_active_iter (box, &iter);
 				user_added = TRUE;
 			}
