@@ -15,7 +15,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * Copyright (C) 2008 - 2011 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.
+ * Copyright (C) 2008 - 2013 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.
  *
  **************************************************************************/
 
@@ -47,14 +47,22 @@
 #define AUTH_TAG "auth "
 #define AUTH_USER_PASS_TAG "auth-user-pass"
 #define CA_TAG "ca "
+#define CA_BLOB_START_TAG "<ca>"
+#define CA_BLOB_END_TAG "</ca>"
 #define CERT_TAG "cert "
+#define CERT_BLOB_START_TAG "<cert>"
+#define CERT_BLOB_END_TAG "</cert>"
 #define CIPHER_TAG "cipher "
+#define KEYSIZE_TAG "keysize "
 #define CLIENT_TAG "client"
 #define COMP_TAG "comp-lzo"
 #define DEV_TAG "dev "
+#define DEV_TYPE_TAG "dev-type "
 #define FRAGMENT_TAG "fragment "
 #define IFCONFIG_TAG "ifconfig "
 #define KEY_TAG "key "
+#define KEY_BLOB_START_TAG "<key>"
+#define KEY_BLOB_END_TAG "</key>"
 #define MSSFIX_TAG "mssfix"
 #define PKCS12_TAG "pkcs12 "
 #define PORT_TAG "port "
@@ -70,6 +78,7 @@
 #define TLS_AUTH_TAG "tls-auth "
 #define TLS_CLIENT_TAG "tls-client"
 #define TLS_REMOTE_TAG "tls-remote "
+#define REMOTE_CERT_TLS_TAG "remote-cert-tls "
 #define TUNMTU_TAG "tun-mtu "
 
 
@@ -153,6 +162,99 @@ handle_path_item (const char *line,
 	return TRUE;
 }
 
+#define CERT_BEGIN  "-----BEGIN CERTIFICATE-----"
+#define CERT_END    "-----END CERTIFICATE-----"
+#define PRIV_KEY_BEGIN  "-----BEGIN PRIVATE KEY-----"
+#define PRIV_KEY_END    "-----END PRIVATE KEY-----"
+
+static gboolean
+handle_blob_item (const char ***line,
+                  const char *key,
+                  NMSettingVPN *s_vpn,
+                  const char *name,
+                  GError **error)
+{
+	gboolean success = FALSE;
+	const char *blob_mark_start, *blob_mark_end;
+	const char *start_tag, *end_tag;
+	char *filename = NULL;
+	char *dirname = NULL;
+	char *path = NULL;
+	GString *in_file = NULL;
+	const char **p;
+
+	if (!strcmp (key, NM_OPENVPN_KEY_CA)) {
+		start_tag = CA_BLOB_START_TAG;
+		end_tag = CA_BLOB_END_TAG;
+		blob_mark_start = CERT_BEGIN;
+		blob_mark_end = CERT_END;
+	} else if (!strcmp (key, NM_OPENVPN_KEY_CERT)) {
+		start_tag = CERT_BLOB_START_TAG;
+		end_tag = CERT_BLOB_END_TAG;
+		blob_mark_start = CERT_BEGIN;
+		blob_mark_end = CERT_END;
+	} else if (!strcmp (key, NM_OPENVPN_KEY_KEY)) {
+		start_tag = KEY_BLOB_START_TAG;
+		end_tag = KEY_BLOB_END_TAG;
+		blob_mark_start = PRIV_KEY_BEGIN;
+		blob_mark_end = PRIV_KEY_END;
+	} else
+		g_return_val_if_reached (FALSE);
+
+	p = *line;
+	if (strncmp (*p, start_tag, strlen (start_tag)))
+		goto finish;
+	p++;
+	if (strcmp (*p, blob_mark_start))
+		goto finish;
+	p++;
+
+	in_file = g_string_new (NULL);
+
+	while (*p && strcmp (*p, blob_mark_end)) {
+		g_string_append (in_file, *p);
+		g_string_append_c (in_file, '\n');
+		p++;
+	}
+	if (!*p || strncmp (*(p+1), end_tag, strlen (end_tag)))
+		goto finish;
+	p++;
+
+	/* Construct file name to write the data in */
+	filename = g_strdup_printf ("%s-%s.pem", name, key);
+	dirname = g_build_filename (g_get_home_dir (), ".cert", NULL);
+	path = g_build_filename (dirname, filename, NULL);
+
+	/* Check that dirname exists and is a directory, otherwise create it */
+	if (!g_file_test (dirname, G_FILE_TEST_IS_DIR)) {
+		if (!g_file_test (dirname, G_FILE_TEST_EXISTS)) {
+			if (mkdir (dirname, 0755) < 0)
+				goto finish;  /* dirname could not be created */
+		} else
+			goto finish;  /* dirname is not a directory */
+	}
+
+	/* Write the new file */
+	g_string_prepend_c (in_file, '\n');
+	g_string_prepend (in_file, blob_mark_start);
+	g_string_append_printf (in_file, "%s", blob_mark_end);
+	success = g_file_set_contents (path, in_file->str, -1, error);
+	if (!success)
+		goto finish;
+
+	nm_setting_vpn_add_data_item (s_vpn, key, path);
+
+finish:
+	line = &p;
+	g_free (filename);
+	g_free (dirname);
+	g_free (path);
+	if (in_file)
+		g_string_free (in_file, TRUE);
+	return success;
+
+}
+
 static char **
 get_args (const char *line, int *nitems)
 {
@@ -302,7 +404,6 @@ do_import (const char *path, char **lines, GError **error)
 	if (last_dot)
 		*last_dot = '\0';
 	g_object_set (s_con, NM_SETTING_CONNECTION_ID, basename, NULL);
-	g_free (basename);
 
 	for (line = lines; *line; line++) {
 		char *comment, **items = NULL, *leftover = NULL;
@@ -323,12 +424,21 @@ do_import (const char *path, char **lines, GError **error)
 		if (!strncmp (*line, DEV_TAG, strlen (DEV_TAG))) {
 			items = get_args (*line + strlen (DEV_TAG), &nitems);
 			if (nitems == 1) {
-				if (g_str_has_prefix (items[0], "tun")) {
-					/* ignore; default is tun */
-				} else if (g_str_has_prefix (items[0], "tap"))
-					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_TAP_DEV, "yes");
+				nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_DEV, items[0]);
+			} else
+				g_warning ("%s: invalid number of arguments in option '%s'", __func__, *line);
+
+			g_strfreev (items);
+			continue;
+		}
+
+		if (!strncmp (*line, DEV_TYPE_TAG, strlen (DEV_TYPE_TAG))) {
+			items = get_args (*line + strlen (DEV_TYPE_TAG), &nitems);
+			if (nitems == 1) {
+				if (!strcmp (items[0], "tun") || !strcmp (items[0], "tap"))
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_DEV_TYPE, items[0]);
 				else
-					g_warning ("%s: unknown %s option '%s'", __func__, DEV_TAG, *line);
+					g_warning ("%s: unknown %s option '%s'", __func__, DEV_TYPE_TAG, *line);
 			} else
 				g_warning ("%s: invalid number of arguments in option '%s'", __func__, *line);
 
@@ -430,9 +540,7 @@ do_import (const char *path, char **lines, GError **error)
 
 		if (   !strncmp (*line, HTTP_PROXY_RETRY_TAG, strlen (HTTP_PROXY_RETRY_TAG))
 		    || !strncmp (*line, SOCKS_PROXY_RETRY_TAG, strlen (SOCKS_PROXY_RETRY_TAG))) {
-			nm_setting_vpn_add_data_item (s_vpn,
-			                              g_strdup (NM_OPENVPN_KEY_PROXY_RETRY),
-			                              g_strdup ("yes"));
+			nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_RETRY, "yes");
 			continue;
 		}
 
@@ -500,7 +608,10 @@ do_import (const char *path, char **lines, GError **error)
 		if (!strncmp (*line, REMOTE_TAG, strlen (REMOTE_TAG))) {
 			items = get_args (*line + strlen (REMOTE_TAG), &nitems);
 			if (nitems >= 1 && nitems <= 3) {
-				nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE, items[0]);
+				const char *prev = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE);
+				char *new_remote = g_strdup_printf ("%s%s%s", prev ? prev : "", prev ? ", " : "", items[0]);
+				nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE, new_remote);
+				g_free (new_remote);
 				have_remote = TRUE;
 
 				if (nitems >= 2) {
@@ -561,6 +672,15 @@ do_import (const char *path, char **lines, GError **error)
 		if (handle_path_item (*line, KEY_TAG, NM_OPENVPN_KEY_KEY, s_vpn, default_path, NULL))
 			continue;
 
+		if (handle_blob_item ((const char ***)&line, NM_OPENVPN_KEY_CA, s_vpn, basename, NULL))
+			continue;
+
+		if (handle_blob_item ((const char ***)&line, NM_OPENVPN_KEY_CERT, s_vpn, basename, NULL))
+			continue;
+
+		if (handle_blob_item ((const char ***)&line, NM_OPENVPN_KEY_KEY, s_vpn, basename, NULL))
+			continue;
+
 		if (handle_path_item (*line, SECRET_TAG, NM_OPENVPN_KEY_STATIC_KEY,
 		                      s_vpn, default_path, &leftover)) {
 			handle_direction ("secret",
@@ -593,6 +713,25 @@ do_import (const char *path, char **lines, GError **error)
 			continue;
 		}
 
+		if (!strncmp (*line, KEYSIZE_TAG, strlen (KEYSIZE_TAG))) {
+			items = get_args (*line + strlen (KEYSIZE_TAG), &nitems);
+			if (nitems == 1) {
+				glong key_size;
+
+				errno = 0;
+				key_size = strtol (items[0], NULL, 10);
+				if ((errno == 0) && (key_size > 0) && (key_size <= 65535)) {
+					tmp = g_strdup_printf ("%d", (guint32) key_size);
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_KEYSIZE, tmp);
+					g_free (tmp);
+				} else
+					g_warning ("%s: invalid key size in option '%s'", __func__, *line);
+			} else
+				g_warning ("%s: invalid number of arguments in option '%s'", __func__, *line);
+			g_strfreev (items);
+			continue;
+		}
+
 		if (!strncmp (*line, TLS_REMOTE_TAG, strlen (TLS_REMOTE_TAG))) {
 			char *unquoted = unquote (*line + strlen (TLS_REMOTE_TAG), NULL);
 
@@ -605,6 +744,20 @@ do_import (const char *path, char **lines, GError **error)
 			continue;
 		}
 
+		if (!strncmp (*line, REMOTE_CERT_TLS_TAG, strlen (REMOTE_CERT_TLS_TAG))) {
+			items = get_args (*line + strlen (REMOTE_CERT_TLS_TAG), &nitems);
+			if (nitems == 1) {
+				if (   !strcmp (items[0], NM_OPENVPN_REM_CERT_TLS_CLIENT)
+				    || !strcmp (items[0], NM_OPENVPN_REM_CERT_TLS_SERVER)) {
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_CERT_TLS, items[0]);
+				} else
+					g_warning ("%s: unknown %s option '%s'", __func__, REMOTE_CERT_TLS_TAG, *line);
+			}
+
+			g_strfreev (items);
+			continue;
+		}
+
 		if (!strncmp (*line, IFCONFIG_TAG, strlen (IFCONFIG_TAG))) {
 			items = get_args (*line + strlen (IFCONFIG_TAG), &nitems);
 			if (nitems == 2) {
@@ -698,6 +851,7 @@ do_import (const char *path, char **lines, GError **error)
 	}
 
 	g_free (default_path);
+	g_free (basename);
 
 	if (connection)
 		nm_connection_add_setting (connection, NM_SETTING (s_vpn));
@@ -714,7 +868,8 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	NMSettingVPN *s_vpn;
 	FILE *f;
 	const char *value;
-	const char *gateway = NULL;
+	const char *gateways = NULL;
+	char **gw_list, **gw_iter;
 	const char *cipher = NULL;
 	const char *cacert = NULL;
 	const char *connection_type = NULL;
@@ -726,14 +881,19 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	const char *local_ip = NULL;
 	const char *remote_ip = NULL;
 	const char *tls_remote = NULL;
+	const char *remote_cert_tls = NULL;
 	const char *tls_auth = NULL;
 	const char *tls_auth_dir = NULL;
+	const char *device = NULL;
+	const char *device_type = NULL;
+	const char *device_default = "tun";
 	gboolean success = FALSE;
-	gboolean device_tun = TRUE;
 	gboolean proto_udp = TRUE;
 	gboolean use_lzo = FALSE;
 	gboolean reneg_exists = FALSE;
 	guint32 reneg = 0;
+	gboolean keysize_exists = FALSE;
+	guint32 keysize = 0;
 	const char *proxy_type = NULL;
 	const char *proxy_server = NULL;
 	const char *proxy_port = NULL;
@@ -754,7 +914,7 @@ do_export (const char *path, NMConnection *connection, GError **error)
 
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE);
 	if (value && strlen (value))
-		gateway = value;
+		gateways = value;
 	else {
 		g_set_error (error, 0, 0, "connection was incomplete (missing gateway)");
 		goto done;
@@ -813,9 +973,18 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (value && !strcmp (value, "yes"))
 		proto_udp = FALSE;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV);
+	if (value && strlen (value))
+		device = value;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV_TYPE);
+	if (value && strlen (value))
+		device_type = value;
+
+	/* Read legacy 'tap-dev' property for backwards compatibility. */
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TAP_DEV);
 	if (value && !strcmp (value, "yes"))
-		device_tun = FALSE;
+		device_default = "tap";
 
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMP_LZO);
 	if (value && !strcmp (value, "yes"))
@@ -825,6 +994,12 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (value && strlen (value))
 		cipher = value;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEYSIZE);
+	if (value && strlen (value)) {
+		keysize_exists = TRUE;
+		keysize = strtol (value, NULL, 10);
+	}
+
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_LOCAL_IP);
 	if (value && strlen (value))
 		local_ip = value;
@@ -841,13 +1016,24 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (value && strlen (value))
 		tls_auth_dir = value;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_CERT_TLS);
+	if (value && strlen (value))
+		remote_cert_tls = value;
+
 	/* Advanced values end */
 
 	fprintf (f, "client\n");
-	fprintf (f, "remote %s%s%s\n",
-	         gateway,
-	         port ? " " : "",
-	         port ? port : "");
+
+	gw_list = g_strsplit_set (gateways, " ,", 0);
+	for (gw_iter = gw_list; gw_iter && *gw_iter; gw_iter++) {
+		if (**gw_iter == '\0')
+			continue;
+		fprintf (f, "remote %s%s%s\n",
+		         *gw_iter,
+		         port ? " " : "",
+		         port ? port : "");
+	}
+	g_strfreev (gw_list);
 
 	/* Handle PKCS#12 (all certs are the same file) */
 	if (   cacert && user_cert && private_key
@@ -882,6 +1068,9 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (cipher)
 		fprintf (f, "cipher %s\n", cipher);
 
+	if (keysize_exists)
+		fprintf (f, "keysize %d\n", keysize);
+
 	if (use_lzo)
 		fprintf (f, "comp-lzo yes\n");
 
@@ -899,7 +1088,9 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (value && strlen (value))
 		fprintf (f, FRAGMENT_TAG " %d\n", (int) strtol (value, NULL, 10));
 
-	fprintf (f, "dev %s\n", device_tun ? "tun" : "tap");
+	fprintf (f, "dev %s\n", device ? device : (device_type ? device_type : device_default));
+	if (device_type)
+		fprintf (f, "dev-type %s\n", device_type);
 	fprintf (f, "proto %s\n", proto_udp ? "udp" : "tcp");
 
 	if (local_ip && remote_ip)
@@ -910,6 +1101,9 @@ do_export (const char *path, NMConnection *connection, GError **error)
 		if (tls_remote)
 			fprintf (f,"tls-remote \"%s\"\n", tls_remote);
 
+		if (remote_cert_tls)
+			fprintf (f,"remote-cert-tls %s\n", remote_cert_tls);
+
 		if (tls_auth) {
 			fprintf (f, "tls-auth %s%s%s\n",
 			         tls_auth,
